@@ -19,90 +19,86 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 """
 
 import sys, os
-from optparse import OptionParser, make_option
-
-from stgit.utils import *
-from stgit import stack, git, gitmergeonefile
-from stgit.version import version
-from stgit.config import config
-from stgit.commands.common import *
-
-# The commands
-import stgit.commands.add
-import stgit.commands.applied
-import stgit.commands.branch
-import stgit.commands.delete
-import stgit.commands.diff
-import stgit.commands.clean
-import stgit.commands.clone
-import stgit.commands.commit
-import stgit.commands.export
-import stgit.commands.files
-import stgit.commands.float
-import stgit.commands.fold
-import stgit.commands.goto
-import stgit.commands.id
-import stgit.commands.imprt
-import stgit.commands.init
-import stgit.commands.log
-import stgit.commands.mail
-import stgit.commands.new
-import stgit.commands.patches
-import stgit.commands.pick
-import stgit.commands.pop
-import stgit.commands.pull
-import stgit.commands.push
-import stgit.commands.refresh
-import stgit.commands.rename
-import stgit.commands.resolved
-import stgit.commands.rm
-import stgit.commands.series
-import stgit.commands.show
-import stgit.commands.status
-import stgit.commands.top
-import stgit.commands.unapplied
-import stgit.commands.uncommit
+from optparse import OptionParser
 
+import stgit.commands
 
 #
 # The commands map
 #
-commands = {
-    'add':      stgit.commands.add,
-    'applied':  stgit.commands.applied,
-    'branch':   stgit.commands.branch,
-    'delete':   stgit.commands.delete,
-    'diff':     stgit.commands.diff,
-    'clean':    stgit.commands.clean,
-    'clone':    stgit.commands.clone,
-    'commit':   stgit.commands.commit,
-    'export':   stgit.commands.export,
-    'files':    stgit.commands.files,
-    'float':    stgit.commands.float,
-    'fold':     stgit.commands.fold,
-    'goto':     stgit.commands.goto,
-    'id':       stgit.commands.id,
-    'import':   stgit.commands.imprt,
-    'init':     stgit.commands.init,
-    'log':      stgit.commands.log,
-    'mail':     stgit.commands.mail,
-    'new':      stgit.commands.new,
-    'patches':  stgit.commands.patches,
-    'pick':     stgit.commands.pick,
-    'pop':      stgit.commands.pop,
-    'pull':     stgit.commands.pull,
-    'push':     stgit.commands.push,
-    'refresh':  stgit.commands.refresh,
-    'rename':   stgit.commands.rename,
-    'resolved': stgit.commands.resolved,
-    'rm':       stgit.commands.rm,
-    'series':   stgit.commands.series,
-    'show':     stgit.commands.show,
-    'status':   stgit.commands.status,
-    'top':      stgit.commands.top,
-    'unapplied':stgit.commands.unapplied,
-    'uncommit': stgit.commands.uncommit,
-    }
+class Commands(dict):
+    """Commands class. It performs on-demand module loading
+    """
+    def canonical_cmd(self, key):
+        """Return the canonical name for a possibly-shortenned
+        command name.
+        """
+        candidates = [cmd for cmd in self.keys() if cmd.startswith(key)]
+
+        if not candidates:
+            print >> sys.stderr, 'Unknown command: %s' % key
+            print >> sys.stderr, '  Try "%s help" for a list of ' \
+                  'supported commands' % prog
+            sys.exit(1)
+        elif len(candidates) > 1:
+            print >> sys.stderr, 'Ambiguous command: %s' % key
+            print >> sys.stderr, '  Candidates are: %s' \
+                  % ', '.join(candidates)
+            sys.exit(1)
+
+        return candidates[0]
+        
+    def __getitem__(self, key):
+        """Return the command python module name based.
+        """
+        global prog
+
+        cmd_mod = self.get(key) or self.get(self.canonical_cmd(key))
+            
+        __import__('stgit.commands.' + cmd_mod)
+        return getattr(stgit.commands, cmd_mod)
+
+commands = Commands({
+    'add':              'add',
+    'applied':          'applied',
+    'assimilate':       'assimilate',
+    'branch':           'branch',
+    'delete':           'delete',
+    'diff':             'diff',
+    'clean':            'clean',
+    'clone':            'clone',
+    'commit':           'commit',
+    'export':           'export',
+    'files':            'files',
+    'float':            'float',
+    'fold':             'fold',
+    'goto':             'goto',
+    'hide':             'hide',
+    'id':               'id',
+    'import':           'imprt',
+    'init':             'init',
+    'log':              'log',
+    'mail':             'mail',
+    'new':              'new',
+    'patches':          'patches',
+    'pick':             'pick',
+    'pop':              'pop',
+    'pull':             'pull',
+    'push':             'push',
+    'rebase':           'rebase',
+    'refresh':          'refresh',
+    'rename':           'rename',
+    'resolved':         'resolved',
+    'rm':               'rm',
+    'series':           'series',
+    'show':             'show',
+    'status':           'status',
+    'sync':             'sync',
+    'top':              'top',
+    'unapplied':        'unapplied',
+    'uncommit':         'uncommit',
+    'unhide':           'unhide'
+    })
 
 # classification: repository, stack, patch, working copy
 repocommands = (
@@ -113,17 +109,21 @@ repocommands = (
     )
 stackcommands = (
     'applied',
+    'assimilate',
     'clean',
     'commit',
     'float',
     'goto',
+    'hide',
     'init',
     'pop',
     'push',
+    'rebase',
     'series',
     'top',
     'unapplied',
-    'uncommit'
+    'uncommit',
+    'unhide'
     )
 patchcommands = (
     'delete',
@@ -137,7 +137,8 @@ patchcommands = (
     'pick',
     'refresh',
     'rename',
-    'show'
+    'show',
+    'sync'
     )
 wccommands = (
     'add',
@@ -192,10 +193,12 @@ def print_help():
 def main():
     """The main function
     """
+    global prog
+
     prog = os.path.basename(sys.argv[0])
 
     if len(sys.argv) < 2:
-        print >> sys.stderr, 'Unknown command'
+        print >> sys.stderr, 'usage: %s <command>' % prog
         print >> sys.stderr, \
               '  Try "%s --help" for a list of supported commands' % prog
         sys.exit(1)
@@ -203,15 +206,15 @@ def main():
     cmd = sys.argv[1]
 
     if cmd in ['-h', '--help']:
-        if len(sys.argv) == 3 and sys.argv[2] in commands:
-            cmd = sys.argv[2]
+        if len(sys.argv) >= 3:
+            cmd = commands.canonical_cmd(sys.argv[2])
             sys.argv[2] = '--help'
         else:
             print_help()
             sys.exit(0)
     if cmd == 'help':
         if len(sys.argv) == 3 and not sys.argv[2] in ['-h', '--help']:
-            cmd = sys.argv[2]
+            cmd = commands.canonical_cmd(sys.argv[2])
             if not cmd in commands:
                 print >> sys.stderr, '%s help: "%s" command unknown' \
                       % (prog, cmd)
@@ -221,12 +224,13 @@ def main():
             command = commands[cmd]
             parser = OptionParser(usage = command.usage,
                                   option_list = command.options)
-            parser.print_help()
+            from pydoc import pager
+            pager(parser.format_help())
         else:
-            print 'usage: %s help <command>' % prog
-
+            print_help()
         sys.exit(0)
     if cmd in ['-v', '--version', 'version']:
+        from stgit.version import version
         print 'Stacked GIT %s' % version
         os.system('git --version')
         print 'Python version %s' % sys.version
@@ -234,35 +238,53 @@ def main():
     if cmd in ['copyright']:
         print __copyright__
         sys.exit(0)
-    if not cmd in commands:
-        print >> sys.stderr, 'Unknown command: %s' % cmd
-        print >> sys.stderr, '  Try "%s help" for a list of supported ' \
-              'commands' % prog
-        sys.exit(1)
 
     # re-build the command line arguments
-    sys.argv[0] += ' %s' % cmd
+    sys.argv[0] += ' %s' % commands.canonical_cmd(cmd)
     del(sys.argv[1])
 
     command = commands[cmd]
     usage = command.usage.split('\n')[0].strip()
     parser = OptionParser(usage = usage, option_list = command.options)
     options, args = parser.parse_args()
+
+    # These modules are only used from this point onwards and do not
+    # need to be imported earlier
+    from stgit.config import config_setup
+    from ConfigParser import ParsingError, NoSectionError
+    from stgit.stack import Series, StackException
+    from stgit.git import GitException
+    from stgit.commands.common import CmdException
+    from stgit.gitmergeonefile import GitMergeException
+
     try:
+        debug_level = int(os.environ['STGIT_DEBUG_LEVEL'])
+    except KeyError:
+        debug_level = 0
+    except ValueError:
+        print >> sys.stderr, 'Invalid STGIT_DEBUG_LEVEL environment variable'
+        sys.exit(1)
+
+    try:
+        config_setup()
+
         # 'clone' doesn't expect an already initialised GIT tree. A Series
         # object will be created after the GIT tree is cloned
         if cmd != 'clone':
             if hasattr(options, 'branch') and options.branch:
-                command.crt_series = stack.Series(options.branch)
+                command.crt_series = Series(options.branch)
             else:
-                command.crt_series = stack.Series()
+                command.crt_series = Series()
             stgit.commands.common.crt_series = command.crt_series
 
         command.func(parser, options, args)
-    except (IOError, CmdException, stack.StackException, git.GitException,
-            gitmergeonefile.GitMergeException), err:
+    except (IOError, ParsingError, NoSectionError, CmdException,
+            StackException, GitException, GitMergeException), err:
         print >> sys.stderr, '%s %s: %s' % (prog, cmd, err)
-        sys.exit(2)
+        if debug_level:
+            raise
+        else:
+            sys.exit(2)
     except KeyboardInterrupt:
         sys.exit(1)
 
